"""Async GeoTIFF and [Cloud-Optimized GeoTIFF][cogeo] (COG) reader for Python.

[cogeo]: https://cogeo.org/
"""

from . import exceptions, utils
from ._array import RasterArray
from ._gdal_metadata import BandStatistics
from ._geotiff import GeoTIFF
from ._overview import Overview
from ._tile import Tile
from ._transform import BoundingBox
from ._version import __version__
from ._windows import Window

__all__ = [
    "BandStatistics",
    "BoundingBox",
    "GeoTIFF",
    "Overview",
    "RasterArray",
    "Tile",
    "Window",
    "__version__",
    "exceptions",
    "utils",
]



from __future__ import annotations

import warnings
from dataclasses import dataclass, field
from types import MappingProxyType
from typing import TYPE_CHECKING, Self

import numpy as np
from async_tiff import TIFF
from async_tiff.enums import Compression as AsyncTiffCompression
from async_tiff.enums import (
    PhotometricInterpretation as AsyncTiffPhotometricInterpretation,
)
from async_tiff.enums import (
    PlanarConfiguration,
    SampleFormat,
)

from async_geotiff._colorinterp import infer_color_interpretation
from async_geotiff._crs import crs_from_geo_keys
from async_geotiff._fetch import FetchTileMixin
from async_geotiff._gdal_metadata import (
    BandStatistics,
    GDALMetadata,
    parse_gdal_metadata,
)
from async_geotiff._overview import Overview
from async_geotiff._read import ReadMixin
from async_geotiff._tile import TiledMixin
from async_geotiff._transform import TransformMixin, create_transform
from async_geotiff.colormap import Colormap
from async_geotiff.enums import (
    ColorInterp,
    Compression,
    Interleaving,
    PhotometricInterpretation,
)

if TYPE_CHECKING:
    from collections.abc import Mapping

    from affine import Affine
    from async_tiff import GeoKeyDirectory, ImageFileDirectory, ObspecInput
    from pyproj.crs import CRS


@dataclass(frozen=True, init=False, kw_only=True, repr=False)
class GeoTIFF(ReadMixin, FetchTileMixin, TiledMixin, TransformMixin):
    """A class representing a GeoTIFF image."""

    _crs: CRS | None = None
    """A cached CRS instance.

    We don't use functools.cached_property on the `crs` attribute because of typing
    issues.
    """

    _tiff: TIFF
    """The underlying async-tiff TIFF instance that we wrap.
    """

    _primary_ifd: ImageFileDirectory = field(init=False)
    """The primary (first) IFD of the GeoTIFF.

    Some tags, like most geo tags, only exist on the primary IFD.
    """

    _mask_ifd: ImageFileDirectory | None = None
    """The mask IFD of the full-resolution GeoTIFF, if any.
    """

    _gkd: GeoKeyDirectory = field(init=False)
    """The GeoKeyDirectory of the primary IFD.
    """

    _overviews: list[Overview] = field(init=False)
    """A list of overviews for the GeoTIFF.
    """

    _gdal_metadata: GDALMetadata | None = None
    """The metadata extracted from the GDALMetadata TIFF tag, if any."""

    @property
    def _alpha_band_idx(self) -> int | None:
        """The index of the alpha band, if any."""
        # TODO: when we support fetching partial bands, we need to check if the alpha
        # band is included in the bands we've fetched.
        # https://github.com/developmentseed/async-geotiff/issues/113
        alpha_band_idxs = [
            i
            for i, colorinterp in enumerate(self._geotiff.colorinterp)
            if colorinterp == ColorInterp.ALPHA
        ]
        if len(alpha_band_idxs) > 1:
            raise ValueError("Multiple alpha bands are not supported")

        return alpha_band_idxs[0] if alpha_band_idxs else None

    @property
    def _ifd(self) -> ImageFileDirectory:
        """An alias for the primary IFD to satisfy _fetch protocol."""
        return self._primary_ifd

    @property
    def _geotiff(self) -> GeoTIFF:
        """An alias for self to satisfy _fetch protocol."""
        return self

    def __init__(self, tiff: TIFF) -> None:
        """Create a GeoTIFF from an existing TIFF instance."""
        first_ifd = tiff.ifds[0]
        gkd = first_ifd.geo_key_directory

        # Validate that this is indeed a GeoTIFF
        if gkd is None:
            raise ValueError("TIFF does not contain GeoTIFF keys")

        if len(tiff.ifds) == 0:
            raise ValueError("TIFF does not contain any IFDs")

        # We use object.__setattr__ because the dataclass is frozen
        object.__setattr__(self, "_tiff", tiff)
        object.__setattr__(self, "_primary_ifd", first_ifd)
        object.__setattr__(self, "_gkd", gkd)
        object.__setattr__(
            self,
            "_gdal_metadata",
            parse_gdal_metadata(
                first_ifd.gdal_metadata,
                count=first_ifd.samples_per_pixel,
            ),
        )

        for ifd in tiff.ifds:
            if ifd.strip_byte_counts is not None or ifd.strip_offsets is not None:
                # This is not an error so that we can test metadata parsing for some
                # striped tiffs against rasterio.
                # Additionally some "single tile" tiffs are actually stored with strip
                # tags instead of tile tags
                warnings.warn(
                    "Striped GeoTIFFs are not supported.",
                    UserWarning,
                    stacklevel=2,
                )

        # Separate data IFDs and mask IFDs (skip the primary IFD at index 0)
        # Data IFDs are indexed by (width, height) for matching with masks
        data_ifds: dict[tuple[int, int], ImageFileDirectory] = {}
        mask_ifds: dict[tuple[int, int], ImageFileDirectory] = {}

        for ifd in tiff.ifds[1:]:
            dims = (ifd.image_width, ifd.image_height)
            if is_mask_ifd(ifd):
                mask_ifds[dims] = ifd
            else:
                data_ifds[dims] = ifd

        # Find and set the mask for the primary IFD (matches primary dimensions)
        if primary_mask_ifd := mask_ifds.get(
            (first_ifd.image_width, first_ifd.image_height),
        ):
            object.__setattr__(self, "_mask_ifd", primary_mask_ifd)

        # Build overviews, sorted by resolution (highest to lowest, i.e., largest first)
        # Sort by width * height descending
        sorted_dims = sorted(data_ifds.keys(), key=lambda d: d[0] * d[1], reverse=True)

        overviews: list[Overview] = []
        for dims in sorted_dims:
            data_ifd = data_ifds[dims]
            mask_ifd = mask_ifds.get(dims)

            ovr = Overview._create(  # noqa: SLF001
                geotiff=self,
                gkd=gkd,
                ifd=data_ifd,
                mask_ifd=mask_ifd,
            )
            overviews.append(ovr)

        object.__setattr__(self, "_overviews", overviews)

    @classmethod
    async def open(
        cls,
        path: str,
        *,
        store: ObspecInput,
        prefetch: int = 32768,
        multiplier: float = 2.0,
    ) -> Self:
        """Open a new GeoTIFF.

        Args:
            path: The path within the store to read from.
            store: The backend to use for data fetching.
            prefetch: The number of initial bytes to read up front.
            multiplier: The multiplier to use for readahead size growth. Must be
                greater than 1.0. For example, for a value of `2.0`, the first metadata
                read will be of size `prefetch`, and then the next read will be of size
                `prefetch * 2`.

        Returns:
            A GeoTIFF instance.

        """
        tiff = await TIFF.open(
            path=path,
            store=store,
            prefetch=prefetch,
            multiplier=multiplier,
        )
        return cls(tiff)

    @property
    def colorinterp(self) -> tuple[ColorInterp, ...]:
        """The color interpretation of each band in index order."""
        interps = list(
            infer_color_interpretation(
                count=self.count,
                photometric=self.photometric,
                extra_samples=self._primary_ifd.extra_samples or [],
            ),
        )

        if gdal_metadata := self._gdal_metadata:
            for band_index, color_interp in gdal_metadata.colorinterp.items():
                interps[band_index - 1] = color_interp

        return tuple(interps)

    @property
    def colormap(self) -> Colormap | None:
        """Return the Colormap stored in the file, if any.

        Returns:
            A Colormap instance if the dataset has a colormap, else None.

        """
        if upstream_colormap := self._primary_ifd.colormap:
            return Colormap(_cmap=upstream_colormap, _nodata=self.nodata)

        return None

    @property
    def compression(self) -> Compression | None:  # noqa: PLR0911
        """The compression algorithm used for the dataset.

        Returns None if the compression is not recognized.
        """
        match self._primary_ifd.compression:
            case AsyncTiffCompression.JPEG:
                return Compression.JPEG
            case AsyncTiffCompression.LZW:
                return Compression.LZW
            case AsyncTiffCompression.PackBits:
                return Compression.PACKBITS
            case AsyncTiffCompression.Deflate:
                return Compression.DEFLATE
            case AsyncTiffCompression.Uncompressed:
                return Compression.UNCOMPRESSED
            case AsyncTiffCompression.ZSTD:
                return Compression.ZSTD

        return None

    @property
    def count(self) -> int:
        """The number of raster bands in the full image."""
        return self._primary_ifd.samples_per_pixel

    @property
    def crs(self) -> CRS:
        """The dataset's coordinate reference system."""
        if self._crs is not None:
            return self._crs

        crs = crs_from_geo_keys(self._gkd)

        # We manually manage the cached property here because `@cached_property` messes
        # with type-checking
        object.__setattr__(self, "_crs", crs)

        return crs

    @property
    def dtype(self) -> np.dtype | None:  # noqa: PLR0911, C901
        """The NumPy data type of the image.

        Returns None if the data type is unknown/not supported.
        """
        formats = set(self._primary_ifd.sample_format)
        bits_per_sample = set(self._primary_ifd.bits_per_sample)

        if len(formats) != 1:
            raise ValueError("Mixed sample formats are not supported.")
        if len(bits_per_sample) != 1:
            raise ValueError("Mixed bits per sample are not supported.")

        match formats.pop(), bits_per_sample.pop():
            case (SampleFormat.Uint, 8):
                return np.dtype(np.uint8)
            case (SampleFormat.Uint, 16):
                return np.dtype(np.uint16)
            case (SampleFormat.Uint, 32):
                return np.dtype(np.uint32)
            case (SampleFormat.Uint, 64):
                return np.dtype(np.uint64)
            case (SampleFormat.Float, 32):
                return np.dtype(np.float32)
            case (SampleFormat.Float, 64):
                return np.dtype(np.float64)
            case (SampleFormat.Int, 8):
                return np.dtype(np.int8)
            case (SampleFormat.Int, 16):
                return np.dtype(np.int16)
            case (SampleFormat.Int, 32):
                return np.dtype(np.int32)
            case (SampleFormat.Int, 64):
                return np.dtype(np.int64)

        return None

    @property
    def height(self) -> int:
        """The height (number of rows) of the full image."""
        return self._primary_ifd.image_height

    @property
    def interleaving(self) -> Interleaving:
        """The interleaving scheme of the dataset."""
        planar_configuration = self._primary_ifd.planar_configuration

        if planar_configuration == PlanarConfiguration.Planar:
            return Interleaving.BAND

        return Interleaving.PIXEL

    @property
    def nodata(self) -> float | None:
        """The dataset's single nodata value."""
        nodata = self._primary_ifd.gdal_nodata
        if nodata is None:
            return None

        return float(nodata)

    @property
    def offsets(self) -> tuple[float, ...]:
        """The offset for each band, if any."""
        if gdal_metadata := self._gdal_metadata:
            return gdal_metadata.offsets

        return tuple([0.0] * self.count)

    @property
    def overviews(self) -> list[Overview]:
        """A list of overview levels for the dataset.

        Overviews are reduced-resolution versions of the main image used for faster
        rendering at lower zoom levels.

        This list of overviews is ordered from finest to coarsest resolution. The first
        element of the list is the highest-resolution after the base image.
        """
        return self._overviews

    @property
    def photometric(self) -> PhotometricInterpretation | None:  # noqa: PLR0911
        """The photometric interpretation of the dataset.

        Returns None if the photometric interpretation is not recognized.
        """
        match self._primary_ifd.photometric_interpretation:
            case AsyncTiffPhotometricInterpretation.WhiteIsZero:
                return PhotometricInterpretation.WHITE_IS_ZERO
            case AsyncTiffPhotometricInterpretation.BlackIsZero:
                return PhotometricInterpretation.BLACK_IS_ZERO
            case AsyncTiffPhotometricInterpretation.RGB:
                return PhotometricInterpretation.RGB
            case AsyncTiffPhotometricInterpretation.RGBPalette:
                return PhotometricInterpretation.RGBPALETTE
            case AsyncTiffPhotometricInterpretation.TransparencyMask:
                return PhotometricInterpretation.TRANSPARENCY_MASK
            case AsyncTiffPhotometricInterpretation.CMYK:
                return PhotometricInterpretation.CMYK
            case AsyncTiffPhotometricInterpretation.YCbCr:
                return PhotometricInterpretation.YCBCR
            case AsyncTiffPhotometricInterpretation.CIELab:
                return PhotometricInterpretation.CIELAB

        return None

    @property
    def scales(self) -> tuple[float, ...]:
        """The scale for each band, if any.

        The keys of the mapping are **1-based** band indices to match GDAL's convention.
        """
        if gdal_metadata := self._gdal_metadata:
            return gdal_metadata.scales

        return tuple([1.0] * self.count)

    @property
    def shape(self) -> tuple[int, int]:
        """Get the shape (height, width) of the full image."""
        return (self.height, self.width)

    @property
    def stored_stats(self) -> Mapping[int, BandStatistics] | None:
        """The **pre-existing** statistics for each GeoTIFF band, if available.

        These statistics are extracted from the GDALMetadata TIFF tag and are never
        computed on demand here.

        The keys of the mapping are **1-based** band indices to match GDAL's convention.
        """
        gdal_metadata = self._gdal_metadata

        # For an empty band_statistics dict, return None
        if gdal_metadata is not None and (stats := gdal_metadata.band_statistics):
            return MappingProxyType(stats)

        return None

    @property
    def tile_height(self) -> int:
        """The height in pixels per tile of the image."""
        # A TIFF with a single tile can omit tile_height
        return self._ifd.tile_height or self._ifd.image_height

    @property
    def tile_width(self) -> int:
        """The width in pixels per tile of the image."""
        # A TIFF with a single tile can omit tile_width
        return self._ifd.tile_width or self._ifd.image_width

    @property
    def transform(self) -> Affine:
        """Return the dataset's georeferencing transformation matrix.

        This transform maps pixel row/column coordinates to coordinates in the dataset's
        CRS.
        """
        return create_transform(
            model_tiepoint=self._primary_ifd.model_tiepoint,
            model_pixel_scale=self._primary_ifd.model_pixel_scale,
            model_transformation=self._primary_ifd.model_transformation,
            raster_type=self._gkd.raster_type,
        )

    @property
    def width(self) -> int:
        """The width (number of columns) of the full image."""
        return self._primary_ifd.image_width


def has_geokeys(ifd: ImageFileDirectory) -> bool:
    """Check if an IFD has GeoTIFF keys.

    Args:
        ifd: The IFD to check.

    """
    return ifd.geo_key_directory is not None


def is_mask_ifd(ifd: ImageFileDirectory) -> bool:
    """Check if an IFD is a mask IFD."""
    return (
        ifd.new_subfile_type is not None
        and ifd.new_subfile_type & 4 != 0
        and ifd.photometric_interpretation
        == AsyncTiffPhotometricInterpretation.TransparencyMask
    )



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

from affine import Affine

from async_geotiff._fetch import FetchTileMixin
from async_geotiff._read import ReadMixin
from async_geotiff._tile import TiledMixin
from async_geotiff._transform import TransformMixin

if TYPE_CHECKING:
    from async_tiff import GeoKeyDirectory, ImageFileDirectory
    from pyproj.crs import CRS

    from async_geotiff import GeoTIFF


@dataclass(init=False, frozen=True, kw_only=True, eq=False, repr=False)
class Overview(ReadMixin, FetchTileMixin, TiledMixin, TransformMixin):
    """An overview level of a Cloud-Optimized GeoTIFF image."""

    _geotiff: GeoTIFF
    """A reference to the parent GeoTIFF object.
    """

    _gkd: GeoKeyDirectory
    """The GeoKeyDirectory of the primary IFD.
    """

    _ifd: ImageFileDirectory
    """The IFD for this overview level."""

    _mask_ifd: ImageFileDirectory | None
    """The IFD for the mask associated with this overview level, if any."""

    @classmethod
    def _create(
        cls,
        *,
        geotiff: GeoTIFF,
        gkd: GeoKeyDirectory,
        ifd: ImageFileDirectory,
        mask_ifd: ImageFileDirectory | None,
    ) -> Overview:
        instance = cls.__new__(cls)

        # We use object.__setattr__ because the dataclass is frozen
        object.__setattr__(instance, "_geotiff", geotiff)
        object.__setattr__(instance, "_gkd", gkd)
        object.__setattr__(instance, "_ifd", ifd)
        object.__setattr__(instance, "_mask_ifd", mask_ifd)

        return instance

    @property
    def count(self) -> int:
        """The number of bands in the overview."""
        return self._geotiff.count

    @property
    def crs(self) -> CRS:
        """The coordinate reference system of the overview."""
        return self._geotiff.crs

    @property
    def height(self) -> int:
        """The height of the overview in pixels."""
        return self._ifd.image_height

    @property
    def nodata(self) -> int | float | None:
        """The nodata value for the overview, if any."""
        return self._geotiff.nodata

    @property
    def tile_height(self) -> int:
        """The height in pixels per tile of the overview."""
        tile_height = self._ifd.tile_height
        assert tile_height is not None, "Constructor validated that images are tiled"  # noqa: S101
        return tile_height

    @property
    def tile_width(self) -> int:
        """The width in pixels per tile of the overview."""
        tile_width = self._ifd.tile_width
        assert tile_width is not None, "Constructor validated that images are tiled"  # noqa: S101
        return tile_width

    @property
    def transform(self) -> Affine:
        """The affine transform mapping pixel coordinates to geographic coordinates.

        Returns:
            Affine: The affine transform.

        """
        full_transform = self._geotiff.transform

        overview_width = self.width
        full_width = self._geotiff.width
        overview_height = self.height
        full_height = self._geotiff.height

        scale_x = full_width / overview_width
        scale_y = full_height / overview_height

        return full_transform * Affine.scale(scale_x, scale_y)

    @property
    def width(self) -> int:
        """The width of the overview in pixels."""
        return self._ifd.image_width



# ruff: noqa: PLR2004

from __future__ import annotations

from typing import TYPE_CHECKING

from async_tiff.enums import ExtraSamples

from async_geotiff.enums import ColorInterp, PhotometricInterpretation

if TYPE_CHECKING:
    from collections.abc import Sequence


def infer_color_interpretation(  # noqa: PLR0911
    *,
    count: int,
    photometric: PhotometricInterpretation | None,
    extra_samples: Sequence[ExtraSamples],
) -> tuple[ColorInterp, ...]:
    """Infer colorinterp array based on GeoTIFF metadata."""
    match photometric:
        case None:
            return (ColorInterp.UNDEFINED,) * count
        case PhotometricInterpretation.BLACK_IS_ZERO:
            return (ColorInterp.GRAY,) + (ColorInterp.UNDEFINED,) * (count - 1)
        case PhotometricInterpretation.RGB:
            if count <= 2:
                raise NotImplementedError(
                    "RGB photometric interpretation with fewer than 3 bands "
                    "is not supported.",
                )

            if count == 3:
                return (
                    ColorInterp.RED,
                    ColorInterp.GREEN,
                    ColorInterp.BLUE,
                )

            if count >= 4:
                # Color interpretations for any extra samples
                # Sample = 2 means alpha, otherwise we mark it as undefined
                # https://web.archive.org/web/20240329145321/https://www.awaresystems.be/imaging/tiff/tifftags/extrasamples.html
                extra_colorinterps = [
                    ColorInterp.ALPHA
                    if sample == ExtraSamples.UnassociatedAlpha
                    else ColorInterp.UNDEFINED
                    for sample in extra_samples or []
                ]
                return (
                    ColorInterp.RED,
                    ColorInterp.GREEN,
                    ColorInterp.BLUE,
                    *extra_colorinterps,
                )

        case PhotometricInterpretation.RGBPALETTE:
            return (ColorInterp.PALETTE,)
        case PhotometricInterpretation.CMYK:
            return (
                ColorInterp.CYAN,
                ColorInterp.MAGENTA,
                ColorInterp.YELLOW,
                ColorInterp.BLACK,
            )
        case PhotometricInterpretation.YCBCR:
            return (
                ColorInterp.Y,
                ColorInterp.Cb,
                ColorInterp.Cr,
            )

    raise NotImplementedError



"""Higher-level read utilities for cross-tile operations."""

from __future__ import annotations

import itertools
from typing import TYPE_CHECKING, Protocol

import numpy as np
from affine import Affine

from async_geotiff._array import RasterArray
from async_geotiff._fetch import HasTiffReference
from async_geotiff._windows import Window
from async_geotiff.exceptions import WindowError

if TYPE_CHECKING:
    from collections.abc import Sequence

    from numpy.typing import NDArray

    from async_geotiff._tile import Tile


class CanFetchTiles(HasTiffReference, Protocol):
    """Protocol for objects that can fetch tiles."""

    async def fetch_tiles(
        self,
        xy: Sequence[tuple[int, int]],
    ) -> list[Tile]: ...


class ReadMixin:
    async def read(
        self: CanFetchTiles,
        *,
        window: Window | None = None,
    ) -> RasterArray:
        """Read pixel data for a window region.

        This method fetches all tiles that intersect the given window and
        stitches them together, returning only the pixels within the window.

        Args:
            window: A Window object defining the pixel region to read.
                If None, the entire image is read.

        Returns:
            A RasterArray containing the pixel data for the requested window.

        Raises:
            WindowError: If the window extends outside the image bounds.

        """
        return await read(self, window=window)


async def read(
    self: CanFetchTiles,
    *,
    window: Window | None = None,
) -> RasterArray:
    if isinstance(window, Window):
        win = window
    else:
        win = Window(col_off=0, row_off=0, width=self.width, height=self.height)

    # Most validation occurred in construction of Window; here we just check against
    # image size
    if win.col_off + win.width > self.width or win.row_off + win.height > self.height:
        raise WindowError(
            f"Window extends outside image bounds.\n"
            f"Window: cols={win.col_off}:{win.col_off + win.width}, "
            f"rows={win.row_off}:{win.row_off + win.height}.\n"
            f"Image size: {self.height}x{self.width}",
        )

    # Calculate which tiles we need to fetch
    tile_x_start = win.col_off // self.tile_width
    tile_x_stop = (win.col_off + win.width - 1) // self.tile_width + 1
    tile_y_start = win.row_off // self.tile_height
    tile_y_stop = (win.row_off + win.height - 1) // self.tile_height + 1

    # Build list of tile coordinates
    xy = list(
        itertools.product(
            range(tile_x_start, tile_x_stop),
            range(tile_y_start, tile_y_stop),
        ),
    )

    tiles = await self.fetch_tiles(xy)

    num_bands = tiles[0].array.count
    dtype = tiles[0].array.data.dtype

    # Create output array and mask array
    output_data = np.empty((num_bands, win.height, win.width), dtype=dtype)
    output_mask: NDArray[np.bool_] | None = None
    if self._mask_ifd is not None:
        output_mask = np.ones((win.height, win.width), dtype=np.bool_)

    assemble_tiles(
        tiles=tiles,
        window=win,
        tile_width=self.tile_width,
        tile_height=self.tile_height,
        output_data=output_data,
        output_mask=output_mask,
    )

    window_transform = self.transform * Affine.translation(
        win.col_off,
        win.row_off,
    )

    return RasterArray(
        data=output_data,
        mask=output_mask,
        width=win.width,
        height=win.height,
        count=num_bands,
        transform=window_transform,
        _geotiff=self._geotiff,
        # TODO: when we support fetching partial bands, we need to check if the
        # alpha band is included in the bands we've fetched.
        # https://github.com/developmentseed/async-geotiff/issues/113
        _alpha_band_idx=self._geotiff._alpha_band_idx,  # noqa: SLF001
    )


def assemble_tiles(  # noqa: PLR0913
    *,
    tiles: list[Tile],
    window: Window,
    tile_width: int,
    tile_height: int,
    output_data: NDArray[np.generic],
    output_mask: NDArray[np.bool_] | None,
) -> None:
    """Assemble multiple tiles into output arrays.

    This function copies data from tiles into the appropriate positions
    in the output arrays, handling partial tiles at window boundaries.
    """
    for tile in tiles:
        # Create a window for this tile's position in image coordinates
        tile_window = Window(
            col_off=tile.x * tile_width,
            row_off=tile.y * tile_height,
            width=tile.array.width,
            height=tile.array.height,
        )

        # Calculate the intersection between tile and target window
        overlap = window.intersection(tile_window)

        # Calculate source slice within the tile
        src_col_start = overlap.col_off - tile_window.col_off
        src_col_stop = src_col_start + overlap.width
        src_row_start = overlap.row_off - tile_window.row_off
        src_row_stop = src_row_start + overlap.height

        # Calculate destination slice within the output
        dst_col_start = overlap.col_off - window.col_off
        dst_col_stop = dst_col_start + overlap.width
        dst_row_start = overlap.row_off - window.row_off
        dst_row_stop = dst_row_start + overlap.height

        # Copy data and mask if present
        output_data[
            :,
            dst_row_start:dst_row_stop,
            dst_col_start:dst_col_stop,
        ] = tile.array.data[:, src_row_start:src_row_stop, src_col_start:src_col_stop]

        if output_mask is not None and tile.array.mask is not None:
            output_mask[
                dst_row_start:dst_row_stop,
                dst_col_start:dst_col_stop,
            ] = tile.array.mask[src_row_start:src_row_stop, src_col_start:src_col_stop]



from __future__ import annotations

from typing import TYPE_CHECKING, TypeVar, overload

import numpy as np

if TYPE_CHECKING:
    from numpy.typing import NDArray

    T = TypeVar("T", bound=np.generic)


# Note: it's important for the np.ma.MaskedArray overload to be first
@overload
def reshape_as_image(arr: np.ma.MaskedArray) -> np.ma.MaskedArray: ...
@overload
def reshape_as_image(arr: NDArray[T]) -> NDArray[T]: ...
def reshape_as_image(
    arr: NDArray[T] | np.ma.MaskedArray,
) -> NDArray[T] | np.ma.MaskedArray:
    """Return the source array reshaped into standard image axis ordering.

    Libraries like async-geotiff and rasterio return arrays in `(bands, rows, columns)`
    order, while image processing and visualization software like Lonboard require
    `(rows, columns, bands)` order.

    This function performs this reshaping by swapping the axes order from (bands, rows,
    columns) to (rows, columns, bands).

    For more information, read the [rasterio documentation on interoperability](https://rasterio.readthedocs.io/en/stable/topics/image_processing.html).

    Args:
        arr : array-like of shape (bands, rows, columns)
            image to reshape

    """
    # This is vendored from rasterio:
    # https://github.com/rasterio/rasterio/blob/1e03dc98d7b63b62fe2f5e34eaeb488be38dfcb7/rasterio/plot.py#L227-L241

    # swap the axes order from (bands, rows, columns) to (rows, columns, bands)
    return np.ma.transpose(arr, [1, 2, 0])



from __future__ import annotations

from collections import defaultdict
from dataclasses import dataclass, field

import defusedxml.ElementTree as ET  # noqa: N817

from async_geotiff.enums import ColorInterp


@dataclass
class BandStatistics:
    """Statistics for a single band in a GeoTIFF."""

    max: float | None = None
    """The maximum pixel value in the band."""
    min: float | None = None
    """The minimum pixel value in the band."""
    mean: float | None = None
    """The mean pixel value in the band."""
    std: float | None = None
    """The standard deviation of the pixel values in the band."""
    valid_percent: float | None = None
    """The percentage of valid pixels in the band."""


@dataclass(frozen=True)
class GDALMetadata:
    """Metadata extracted from the GDALMetadata TIFF tag."""

    band_statistics: dict[int, BandStatistics]
    """A mapping of band index to statistics for that band.

    This band index is **1-based**.
    """

    offsets: tuple[float, ...]

    scales: tuple[float, ...]

    colorinterp: dict[int, ColorInterp] = field(default_factory=dict)
    """A mapping of 1-based band index to overridden ColorInterp.

    When present, these values override what would otherwise be inferred
    from the photometric interpretation tag.
    """


LOWER_CASE_COLORINTERP_MAPPING: dict[str, ColorInterp] = {}
for color in ColorInterp:
    LOWER_CASE_COLORINTERP_MAPPING[color.name.lower()] = color


def parse_gdal_metadata(  # noqa: C901
    gdal_metadata: str | None,
    *,
    count: int,
) -> GDALMetadata | None:
    if gdal_metadata is None:
        return None

    root = ET.fromstring(gdal_metadata)

    if root.tag != "GDALMetadata":
        raise ValueError("Not a GDALMetadata XML block")

    band_statistics: defaultdict[int, BandStatistics] = defaultdict(BandStatistics)
    offsets: list[float] = [0.0] * count
    scales: list[float] = [1.0] * count
    colorinterp: dict[int, ColorInterp] = {}

    for elem in root.findall("Item"):
        name = elem.attrib.get("name")
        sample = elem.attrib.get("sample")
        role = elem.attrib.get("role")
        text = elem.text or ""
        match name:
            # Add 1 to get a 1-based band index to match GDAL.
            case "STATISTICS_MAXIMUM" if sample is not None:
                band_statistics[int(sample) + 1].max = float(text)
            case "STATISTICS_MEAN" if sample is not None:
                band_statistics[int(sample) + 1].mean = float(text)
            case "STATISTICS_MINIMUM" if sample is not None:
                band_statistics[int(sample) + 1].min = float(text)
            case "STATISTICS_STDDEV" if sample is not None:
                band_statistics[int(sample) + 1].std = float(text)
            case "STATISTICS_VALID_PERCENT" if sample is not None:
                band_statistics[int(sample) + 1].valid_percent = float(text)
            case "OFFSET" if sample is not None:
                offsets[int(sample)] = float(text)
            case "SCALE" if sample is not None:
                scales[int(sample)] = float(text)
            case "COLORINTERP" if role == "colorinterp" and sample is not None:
                colorinterp[int(sample) + 1] = LOWER_CASE_COLORINTERP_MAPPING[
                    text.lower()
                ]

    return GDALMetadata(
        band_statistics=dict(band_statistics),
        offsets=tuple(offsets),
        scales=tuple(scales),
        colorinterp=colorinterp,
    )



"""Mixin class for coordinate transformation methods."""

from __future__ import annotations

import math
from math import floor
from typing import TYPE_CHECKING, Literal, NamedTuple, Protocol, cast

from affine import Affine

if TYPE_CHECKING:
    from collections.abc import Callable

RASTER_TYPE_PIXEL_IS_POINT = 2


def create_transform(
    *,
    model_tiepoint: list[float] | None,
    model_pixel_scale: list[float] | None,
    model_transformation: list[float] | None,
    raster_type: int | None,
) -> Affine:
    if model_tiepoint is not None and model_pixel_scale is not None:
        affine = create_from_model_tiepoint_and_pixel_scale(
            model_tiepoint,
            model_pixel_scale,
        )

    elif model_transformation is not None:
        affine = create_from_model_transformation(model_transformation)
    else:
        raise ValueError("The image does not have an affine transformation.")

    # Offset transform by half pixel for point-interpreted rasters.
    if raster_type == RASTER_TYPE_PIXEL_IS_POINT:
        affine = affine * Affine.translation(-0.5, -0.5)

    return affine


def create_from_model_tiepoint_and_pixel_scale(
    model_tiepoint: list[float],
    model_pixel_scale: list[float],
) -> Affine:
    x_origin = model_tiepoint[3]
    y_origin = model_tiepoint[4]
    x_resolution = model_pixel_scale[0]
    y_resolution = -model_pixel_scale[1]

    return Affine(x_resolution, 0, x_origin, 0, y_resolution, y_origin)


def create_from_model_transformation(model_transformation: list[float]) -> Affine:
    # ModelTransformation is a 4x4 matrix in row-major order
    # [0  1  2  3 ]   [a  b  0  c]
    # [4  5  6  7 ] = [d  e  0  f]
    # [8  9  10 11]   [0  0  1  0]
    # [12 13 14 15]   [0  0  0  1]
    x_origin = model_transformation[3]
    y_origin = model_transformation[7]
    row_rotation = model_transformation[1]
    col_rotation = model_transformation[4]
    x_resolution = model_transformation[0]
    y_resolution = model_transformation[5]

    return Affine(
        x_resolution,
        row_rotation,
        x_origin,
        col_rotation,
        y_resolution,
        y_origin,
    )


# https://github.com/rasterio/rasterio/blob/7689d9dc83a51b7cc12d1707ce6b011885f61560/rasterio/coords.py#L5-L17
class BoundingBox(NamedTuple):
    """Bounding box, defining extent in cartesian coordinates."""

    left: float
    """Left coordinate"""

    bottom: float
    """Bottom coordinate."""

    right: float
    """Right coordinate"""

    top: float
    """Top coordinate"""


class HasTransform(Protocol):
    """Protocol for objects that have an affine transform."""

    @property
    def height(self) -> int:
        """The height of the image in pixels."""
        ...

    @property
    def width(self) -> int:
        """The width of the image in pixels."""
        ...

    @property
    def transform(self) -> Affine: ...


class TransformMixin:
    """Mixin providing coordinate transformation methods.

    Classes using this mixin must implement HasTransform.
    """

    @property
    def bounds(self: HasTransform) -> BoundingBox:
        """Return the bounds of the dataset in the units of its CRS.

        Returns:
            BoundingBox: The bounding box of the dataset.

        """
        tr = self.transform
        width = self.width
        height = self.height

        # TODO: remove explicit casts with affine v3
        # https://github.com/developmentseed/async-geotiff/issues/123

        # non-rotated transform
        if tr.b == tr.d == 0:
            c0x, c0y = cast("tuple[float, float]", tr * (0, 0))
            c2x, c2y = cast("tuple[float, float]", tr * (width, height))
            return BoundingBox(c0x, c2y, c2x, c0y)

        # Rotated transform
        c0x, c0y = cast("tuple[float, float]", tr * (0, 0))
        c1x, c1y = cast("tuple[float, float]", tr * (0, height))
        c2x, c2y = cast("tuple[float, float]", tr * (width, height))
        c3x, c3y = cast("tuple[float, float]", tr * (width, 0))
        xs = (c0x, c1x, c2x, c3x)
        ys = (c0y, c1y, c2y, c3y)
        return BoundingBox(min(xs), min(ys), max(xs), max(ys))

    def index(
        self: HasTransform,
        x: float,
        y: float,
        op: Callable[[float], int] = floor,
    ) -> tuple[int, int]:
        """Get the (row, col) index of the pixel containing (x, y).

        Args:
            x: x value in coordinate reference system.
            y: y value in coordinate reference system.
            op: Function to convert fractional pixels to whole numbers
                (floor, ceiling, round). Defaults to math.floor.

        Returns:
            (row index, col index)

        """
        inv_transform = ~self.transform
        # Affine * (x, y) returns tuple[float, float] for 2D coordinates
        col_frac, row_frac = inv_transform * (x, y)  # type: ignore[misc]

        return (op(row_frac), op(col_frac))

    @property
    def res(self: HasTransform) -> tuple[float, float]:
        """Return the (width, height) of pixels in the units of its CRS."""
        transform = self.transform

        # For rotated images, resolution is the magnitude of the pixel size
        # calculated from the transform matrix components
        res_x = math.sqrt(transform.a**2 + transform.d**2)
        res_y = math.sqrt(transform.b**2 + transform.e**2)

        return (res_x, res_y)

    def xy(
        self: HasTransform,
        row: int,
        col: int,
        offset: Literal["center", "ul", "ur", "ll", "lr"] = "center",
    ) -> tuple[float, float]:
        """Get the coordinates (x, y) of a pixel at (row, col).

        The pixel's center is returned by default, but a corner can be returned
        by setting `offset` to one of `"ul"`, `"ur"`, `"ll"`, `"lr"`.

        Args:
            row: Pixel row.
            col: Pixel column.
            offset: Determines if the returned coordinates are for the center of the
                pixel or for a corner.

        Returns:
            (x, y) coordinates in the dataset's CRS.

        """
        if offset == "center":
            c = col + 0.5
            r = row + 0.5
        elif offset == "ul":
            c = col
            r = row
        elif offset == "ur":
            c = col + 1
            r = row
        elif offset == "ll":
            c = col
            r = row + 1
        elif offset == "lr":
            c = col + 1
            r = row + 1
        else:
            raise ValueError(f"Invalid offset value: {offset}")

        return self.transform * (c, r)



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, Self

import numpy as np
from async_tiff.enums import PlanarConfiguration
from numpy.ma import MaskedArray

from async_geotiff._transform import TransformMixin

if TYPE_CHECKING:
    from affine import Affine
    from async_tiff import Array
    from numpy.typing import NDArray
    from pyproj.crs import CRS

    from async_geotiff import GeoTIFF


@dataclass(frozen=True, kw_only=True, eq=False)
class RasterArray(TransformMixin):
    """Georeferenced array data from a GeoTIFF."""

    data: NDArray
    """The array data with shape (bands, height, width)."""

    mask: NDArray[np.bool_] | None
    """The mask array with shape (height, width), if any.

    Values of True indicate valid data; False indicates no data.
    """

    width: int
    """The width of the array in pixels."""

    height: int
    """The height of the array in pixels."""

    count: int
    """The number of bands in the array."""

    transform: Affine
    """The affine transform mapping pixel coordinates to geographic coordinates."""

    _alpha_band_idx: int | None
    """The index of the alpha band, if any.

    The alpha band lives in the `data` array but is checked in `as_masked`.
    """

    _geotiff: GeoTIFF
    """A reference to the parent GeoTIFF."""

    @classmethod
    def _create(  # noqa: PLR0913
        cls,
        *,
        data: Array,
        mask: Array | None,
        planar_configuration: PlanarConfiguration,
        transform: Affine,
        geotiff: GeoTIFF,
        alpha_band_idx: int | None,
    ) -> Self:
        """Create a RasterArray from async_tiff data.

        Handles axis reordering to ensure data is always in (bands, height, width)
        order, matching rasterio's convention.
        """
        data_arr = np.asarray(data, copy=False)
        if mask is not None:
            mask_arr = np.asarray(mask, copy=False).astype(np.bool_, copy=False)
            assert mask_arr.ndim == 3  # noqa: S101, PLR2004
            assert mask_arr.shape[2] == 1  # noqa: S101
            # This assumes it's always (height, width, 1)
            mask_arr = np.squeeze(mask_arr, axis=2)
        else:
            mask_arr = None

        assert data_arr.ndim == 3, f"Expected 3D array, got {data_arr.ndim}D"  # noqa: S101, PLR2004

        # async_tiff returns data in the native TIFF order:
        # - Chunky (pixel interleaved): (height, width, bands)
        # - Planar (band interleaved): (bands, height, width)
        # We always want (bands, height, width) to match rasterio.
        if planar_configuration == PlanarConfiguration.Chunky:
            # Transpose from (H, W, C) to (C, H, W)
            data_arr = np.moveaxis(data_arr, -1, 0)

        count, height, width = data_arr.shape

        return cls(
            data=data_arr,
            mask=mask_arr,
            width=width,
            height=height,
            count=count,
            transform=transform,
            _geotiff=geotiff,
            _alpha_band_idx=alpha_band_idx,
        )

    def as_masked(self) -> MaskedArray:
        """Return the data as a masked array using the array mask or nodata value.

        !!! warning
            In a numpy [`MaskedArray`][numpy.ma.MaskedArray], `True`
            indicates invalid (masked) data and `False` indicates valid data.

            This is the inverse convention of a GeoTIFF's mask. The boolean array
            [`RasterArray.mask`][async_geotiff.RasterArray.mask] uses `True` for valid
            data and `False` for invalid data.

        Returns:
            A masked array with the same shape as `data`, where invalid data
                (as indicated by the mask) is masked out.

        """
        if self.mask is not None:
            # MaskedArray constructor does not automatically broadcast inputs.
            # mask must have exactly the same shape as data.
            mask = np.broadcast_to(~self.mask, self.data.shape)
            return MaskedArray(self.data, mask=mask)

        if self.nodata is not None:
            return np.ma.masked_equal(self.data, self.nodata)

        if self._alpha_band_idx is not None:
            alpha_band = self.data[self._alpha_band_idx]
            single_band_mask = alpha_band == np.iinfo(alpha_band.dtype).min
            mask = np.broadcast_to(
                single_band_mask,
                (self.count - 1, self.height, self.width),
            )
            # Rasterio semantics say that the alpha band itself in `data` should be
            # considered valid data.
            mask = np.insert(mask, self._alpha_band_idx, False, axis=0)
            return MaskedArray(self.data, mask=mask)

        return MaskedArray(self.data)

    @property
    def crs(self) -> CRS:
        """The coordinate reference system of the array."""
        return self._geotiff.crs

    @property
    def nodata(self) -> float | None:
        """The nodata value for the array, if any."""
        return self._geotiff.nodata

    @property
    def shape(self) -> tuple[int, int, int]:
        """The shape of the data array as (bands, height, width)."""
        return self.data.shape



"""Window utilities for defining rectangular subsets of rasters."""

from __future__ import annotations

from dataclasses import dataclass

from async_geotiff.exceptions import WindowError


@dataclass(frozen=True, slots=True)
class Window:
    """A rectangular subset of a raster.

    Windows define pixel regions using column/row offsets and dimensions.
    This class is similar to rasterio's [Window][rasterio.windows.Window] but supports
    integer offsets and ranges only.
    """

    col_off: int
    """The column offset (x position of the left edge)."""

    row_off: int
    """The row offset (y position of the top edge)."""

    width: int
    """The width in pixels (number of columns)."""

    height: int
    """The height in pixels (number of rows)."""

    def __post_init__(self) -> None:
        # Validate window dimensions
        if self.col_off < 0 or self.row_off < 0:
            raise WindowError(
                f"Window start indices must be non-negative, "
                f"got col_off={self.col_off}, row_off={self.row_off}",
            )

        if self.width <= 0:
            raise WindowError(f"Window width must be positive, got {self.width}")

        if self.height <= 0:
            raise WindowError(f"Window height must be positive, got {self.height}")

    def __repr__(self) -> str:
        return (
            f"async_geotiff.Window(col_off={self.col_off}, row_off={self.row_off}, "
            f"width={self.width}, height={self.height})"
        )

    def intersection(self, other: Window) -> Window:
        """Compute the intersection with another window.

        Args:
            other: Another Window object.

        Returns:
            A new Window representing the overlapping region.

        Raises:
            WindowError: If windows do not intersect.

        """
        col_off = max(self.col_off, other.col_off)
        row_off = max(self.row_off, other.row_off)
        col_stop = min(self.col_off + self.width, other.col_off + other.width)
        row_stop = min(self.row_off + self.height, other.row_off + other.height)

        width = col_stop - col_off
        height = row_stop - row_off

        if width <= 0 or height <= 0:
            raise WindowError(f"Windows do not intersect: {self} and {other}")

        return Window(col_off=col_off, row_off=row_off, width=width, height=height)



from __future__ import annotations

import math
from dataclasses import dataclass
from typing import TYPE_CHECKING, Protocol

if TYPE_CHECKING:
    from async_geotiff._array import RasterArray


@dataclass(frozen=True, kw_only=True, eq=False)
class Tile:
    """A tile from a GeoTIFF, containing array data and grid position."""

    x: int
    """The tile column index in the GeoTIFF or overview."""

    y: int
    """The tile row index in the GeoTIFF or overview."""

    array: RasterArray
    """The array data for this tile."""


class IsTiled(Protocol):
    """Protocol for objects that are tiled and can provide tile dimensions."""

    @property
    def tile_width(self) -> int:
        """The width of tiles in pixels."""
        ...

    @property
    def tile_height(self) -> int:
        """The height of tiles in pixels."""
        ...

    @property
    def height(self) -> int:
        """The height of the image in pixels."""
        ...

    @property
    def width(self) -> int:
        """The width of the image in pixels."""
        ...

    @property
    def count(self) -> int:
        """The number of bands."""
        ...


class TiledMixin:
    @property
    def block_shapes(self: IsTiled) -> tuple[tuple[int, int], ...]:
        """Block shapes for each band. Each shape is (height, width)."""
        shape = (self.tile_height, self.tile_width)
        return (shape,) * self.count

    @property
    def tile_count(self: IsTiled) -> tuple[int, int]:
        """The number of tiles in the x and y directions."""
        return (
            math.ceil(self.width / self.tile_width),
            math.ceil(self.height / self.tile_height),
        )



"""Generate a Tile Matrix Set from a GeoTIFF file, using [Morecantile].

[Morecantile]: https://developmentseed.org/morecantile/
"""

from __future__ import annotations

from typing import TYPE_CHECKING
from uuid import uuid4

from morecantile.commons import BoundingBox
from morecantile.models import (
    CRS,
    CRSWKT,
    CRSUri,
    TileMatrix,
    TileMatrixSet,
    TMSBoundingBox,
)
from morecantile.utils import meters_per_unit
from pydantic import AnyUrl

if TYPE_CHECKING:
    import pyproj

    from async_geotiff import GeoTIFF

_SCREEN_PIXEL_SIZE = 0.28e-3

__all__ = ["generate_tms"]


def generate_tms(
    geotiff: GeoTIFF,
    *,
    id: str = str(uuid4()),  # noqa: A002
) -> TileMatrixSet:
    """Generate a [Tile Matrix Set] from a GeoTIFF file.

    [Tile Matrix Set]: https://docs.ogc.org/is/17-083r4/17-083r4.html

    Args:
        geotiff: The GeoTIFF file to generate the TMS from.

    Keyword Args:
        id: The ID to assign to the Tile Matrix Set.

    """
    mpu = meters_per_unit(geotiff.crs)

    tile_matrices: list[TileMatrix] = []
    for idx, overview in enumerate(reversed(geotiff.overviews)):
        ovr_tr = overview.transform
        ovr_matrix_width, ovr_matrix_height = overview.tile_count

        tile_matrices.append(
            TileMatrix(
                id=str(idx),
                scaleDenominator=ovr_tr.a * mpu / _SCREEN_PIXEL_SIZE,
                cellSize=ovr_tr.a,
                cornerOfOrigin="bottomLeft" if ovr_tr.e > 0 else "topLeft",
                pointOfOrigin=(ovr_tr.c, ovr_tr.f),
                tileWidth=overview.tile_width,
                tileHeight=overview.tile_height,
                matrixWidth=ovr_matrix_width,
                matrixHeight=ovr_matrix_height,
            ),
        )

    matrix_width, matrix_height = geotiff.tile_count
    tr = geotiff.transform

    # Add the full-resolution level last
    tile_matrices.append(
        TileMatrix(
            id=str(len(geotiff.overviews)),
            scaleDenominator=tr.a * mpu / _SCREEN_PIXEL_SIZE,
            cellSize=tr.a,
            cornerOfOrigin="bottomLeft" if tr.e > 0 else "topLeft",
            pointOfOrigin=(tr.c, tr.f),
            tileWidth=geotiff.tile_width,
            tileHeight=geotiff.tile_height,
            matrixWidth=matrix_width,
            matrixHeight=matrix_height,
        ),
    )

    bbox = BoundingBox(*geotiff.bounds)
    tms_crs = _parse_crs(geotiff.crs)

    return TileMatrixSet(
        title="Generated TMS",
        id=id,
        crs=tms_crs,
        boundingBox=TMSBoundingBox(
            lowerLeft=(bbox.left, bbox.bottom),
            upperRight=(bbox.right, bbox.top),
            crs=tms_crs,
        ),
        tileMatrices=tile_matrices,
    )


def _parse_crs(
    crs: pyproj.CRS,
) -> CRS:
    """Parse a pyproj CRS into a morecantile CRSUri or CRSWKT.

    Args:
        crs: The pyproj CRS to parse.

    """
    if authority_code := crs.to_authority(min_confidence=20):
        authority, code = authority_code
        version = "0"
        # if we have a version number in the authority, split it out
        if "_" in authority:
            authority, version = authority.split("_")

        return CRS(
            CRSUri(
                uri=AnyUrl(
                    f"http://www.opengis.net/def/crs/{authority}/{version}/{code}",
                ),
            ),
        )

    return CRS(CRSWKT(wkt=crs.to_json_dict()))



"""Enums used by async_geotiff."""

from enum import Enum, IntEnum

# ruff: noqa: D101


# https://github.com/rasterio/rasterio/blob/2d79e5f3a00e919ecaa9573adba34a78274ce48c/rasterio/enums.py#L153-L174
class Compression(Enum):
    """Available compression algorithms for GeoTIFFs."""

    JPEG = "JPEG"
    LZW = "LZW"
    PACKBITS = "PACKBITS"
    DEFLATE = "DEFLATE"
    CCITTRLE = "CCITTRLE"
    CCITTFAX3 = "CCITTFAX3"
    CCITTFAX4 = "CCITTFAX4"
    LZMA = "LZMA"
    NONE = "NONE"
    ZSTD = "ZSTD"
    LERC = "LERC"
    LERC_DEFLATE = "LERC_DEFLATE"
    LERC_ZSTD = "LERC_ZSTD"
    WEBP = "WEBP"
    JPEG2000 = "JPEG2000"
    UNCOMPRESSED = "UNCOMPRESSED"


# https://github.com/rasterio/rasterio/blob/2d79e5f3a00e919ecaa9573adba34a78274ce48c/rasterio/enums.py#L177-L182
class Interleaving(Enum):
    """GeoTIFF band interleaving options."""

    PIXEL = "PIXEL"
    BAND = "BAND"
    # TODO: Support GDAL's new TILE interleaving option
    # https://gdal.org/en/stable/drivers/raster/cog.html#general-creation-options


# https://github.com/rasterio/rasterio/blob/2d79e5f3a00e919ecaa9573adba34a78274ce48c/rasterio/enums.py#L192-L200
class PhotometricInterpretation(Enum):
    WHITE_IS_ZERO = "MINISWHITE"
    BLACK_IS_ZERO = "MINISBLACK"
    RGB = "RGB"
    RGBPALETTE = "PALETTE"
    TRANSPARENCY_MASK = "MASK"
    CMYK = "CMYK"
    YCBCR = "YCbCr"
    CIELAB = "CIELAB"


# https://github.com/rasterio/rasterio/blob/04a8620f710814459f2a8dfa2d3d302b66a5408e/rasterio/enums.py#L30-L72
class ColorInterp(IntEnum):
    """Raster band color interpretation."""

    UNDEFINED = 0
    GRAY = 1
    GREY = 1
    PALETTE = 2
    RED = 3
    GREEN = 4
    BLUE = 5
    ALPHA = 6
    HUE = 7
    SATURATION = 8
    LIGHTNESS = 9
    CYAN = 10
    MAGENTA = 11
    YELLOW = 12
    BLACK = 13
    Y = 14
    Cb = 15
    Cr = 16



from __future__ import annotations

import asyncio
from typing import TYPE_CHECKING, Protocol

from affine import Affine

from async_geotiff._array import RasterArray
from async_geotiff._tile import Tile
from async_geotiff._transform import HasTransform

if TYPE_CHECKING:
    from collections.abc import Sequence

    from async_tiff import Array, ImageFileDirectory

    from async_geotiff import GeoTIFF


class HasTiffReference(HasTransform, Protocol):
    """Protocol for objects that hold a TIFF reference and can request tiles."""

    @property
    def _ifd(self) -> ImageFileDirectory:
        """The data IFD for this image (index, IFD)."""
        ...

    @property
    def _mask_ifd(self) -> ImageFileDirectory | None:
        """The mask IFD for this image (index, IFD), if any."""
        ...

    @property
    def _geotiff(self) -> GeoTIFF:
        """The parent GeoTIFF object."""
        ...

    @property
    def tile_height(self) -> int:
        """The height of tiles in pixels."""
        ...

    @property
    def tile_width(self) -> int:
        """The width of tiles in pixels."""
        ...

    @property
    def width(self) -> int:
        """The width of the image in pixels."""
        ...

    @property
    def height(self) -> int:
        """The height of the image in pixels."""
        ...


class FetchTileMixin:
    """Mixin for fetching tiles from a GeoTIFF.

    Classes using this mixin must implement HasTiffReference.
    """

    async def fetch_tile(
        self: HasTiffReference,
        x: int,
        y: int,
        *,
        boundless: bool = True,
    ) -> Tile:
        """Fetch a single tile from the GeoTIFF.

        Args:
            x: The x-coordinate of the tile.
            y: The y-coordinate of the tile.

        Keyword Args:
            boundless: If False, clip edge tiles to the image bounds. Defaults to True.

        Returns:
            A Tile object containing the fetched tile data.

        """
        tile_fut = self._ifd.fetch_tile(x, y)

        mask_data: Array | None = None
        if self._mask_ifd is not None:
            mask_fut = self._mask_ifd.fetch_tile(x, y)
            tile, mask = await asyncio.gather(tile_fut, mask_fut)
            tile_data, mask_data = await asyncio.gather(tile.decode(), mask.decode())
        else:
            tile = await tile_fut
            tile_data = await tile.decode()

        tile_transform = self.transform * Affine.translation(
            x * self.tile_width,
            y * self.tile_height,
        )

        array = RasterArray._create(  # noqa: SLF001
            data=tile_data,
            mask=mask_data,
            planar_configuration=self._ifd.planar_configuration,
            transform=tile_transform,
            geotiff=self._geotiff,
            # TODO: when we support fetching partial bands, we need to check if the
            # alpha band is included in the bands we've fetched.
            # https://github.com/developmentseed/async-geotiff/issues/113
            alpha_band_idx=self._geotiff._alpha_band_idx,  # noqa: SLF001
        )

        if not boundless:
            array = _clip_to_image_bounds(self, x, y, array)

        return Tile(
            x=x,
            y=y,
            array=array,
        )

    async def fetch_tiles(
        self: HasTiffReference,
        xy: Sequence[tuple[int, int]],
        *,
        boundless: bool = True,
    ) -> list[Tile]:
        """Fetch multiple tiles from this overview.

        The benefit of using this method over multiple calls to `fetch_tile` is that
        `fetch_tiles` will coalesce requests for tiles that are stored contiguously on
        disk. The exact behavior depends on how the underlying
        [`store`][async_tiff.ObspecInput] passed to
        [`GeoTIFF.open`][async_geotiff.GeoTIFF.open] implements
        [`get_ranges_async`][obspec.GetRangesAsync].

        Args:
            xy: The (x, y) coordinates of the tiles.

        Keyword Args:
            boundless: If False, clip edge tiles to the image bounds.

        """
        tiles_fut = self._ifd.fetch_tiles(xy)

        decoded_masks: list[Array | None] = [None] * len(xy)
        if self._mask_ifd is not None:
            masks_fut = self._mask_ifd.fetch_tiles(xy)
            tiles, masks = await asyncio.gather(tiles_fut, masks_fut)

            decoded_tile_futs = [tile.decode() for tile in tiles]
            decoded_mask_futs = [mask.decode() for mask in masks]
            decoded_tiles = await asyncio.gather(*decoded_tile_futs)
            decoded_masks = await asyncio.gather(*decoded_mask_futs)
        else:
            tiles = await tiles_fut
            decoded_tiles = await asyncio.gather(*[tile.decode() for tile in tiles])

        final_tiles: list[Tile] = []
        for (x, y), tile_data, mask_data in zip(
            xy,
            decoded_tiles,
            decoded_masks,
            strict=True,
        ):
            tile_transform = self.transform * Affine.translation(
                x * self.tile_width,
                y * self.tile_height,
            )
            array = RasterArray._create(  # noqa: SLF001
                data=tile_data,
                mask=mask_data,
                planar_configuration=self._ifd.planar_configuration,
                transform=tile_transform,
                geotiff=self._geotiff,
                # TODO: when we support fetching partial bands, we need to check if the
                # alpha band is included in the bands we've fetched.
                # https://github.com/developmentseed/async-geotiff/issues/113
                alpha_band_idx=self._geotiff._alpha_band_idx,  # noqa: SLF001
            )

            if not boundless:
                array = _clip_to_image_bounds(self, x, y, array)

            tile = Tile(
                x=x,
                y=y,
                array=array,
            )
            final_tiles.append(tile)

        return final_tiles


def _clip_to_image_bounds(
    self: HasTiffReference,
    x: int,
    y: int,
    array: RasterArray,
) -> RasterArray:
    """Clip a decoded tile array to the valid image bounds.

    Edge tiles in a COG are always encoded at the full tile size, with the
    out-of-bounds region zero-padded. When ``boundless=False`` is requested,
    this function copies only the valid pixel sub-rectangle into a new array,
    returning an Array whose width/height match the actual image content.

    Interior tiles (where the tile fits entirely within the image) are
    returned unchanged.
    """
    clipped_width = min((x + 1) * self.tile_width, self.width) - x * self.tile_width
    clipped_height = min((y + 1) * self.tile_height, self.height) - y * self.tile_height

    if clipped_width == self.tile_width and clipped_height == self.tile_height:
        return array

    # data shape is (bands, height, width)
    clipped_data = array.data[:, :clipped_height, :clipped_width]
    clipped_mask = (
        array.mask[:clipped_height, :clipped_width] if array.mask is not None else None
    )

    return RasterArray(
        data=clipped_data,
        mask=clipped_mask,
        width=clipped_width,
        height=clipped_height,
        count=array.count,
        transform=array.transform,
        _geotiff=array._geotiff,  # noqa: SLF001
        _alpha_band_idx=array._alpha_band_idx,  # noqa: SLF001
    )



"""High-level Colormap class for GeoTIFF colormaps."""

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

import numpy as np

if TYPE_CHECKING:
    from async_tiff import Colormap as AsyncTiffColormap
    from numpy.typing import NDArray


@dataclass(frozen=True, kw_only=True, eq=False)
class Colormap:
    """A representation of a GeoTIFF colormap.

    GeoTIFF colormaps
    """

    _cmap: AsyncTiffColormap
    """The colormap data held in Rust, accessible via the buffer protocol.

    Has shape `(N, 3)` and is of data type uint16.
    """

    _nodata: int | float | None
    """The nodata value from gdal_nodata, if set."""

    def as_array(self, *, dtype: type[np.uint8 | np.uint16] = np.uint8) -> NDArray:
        """Return the colormap as a NumPy array with shape (N, 3) and dtype uint16.

        Each row corresponds to a color entry in the colormap, with columns
        representing the Red, Green, and Blue components respectively.

        This is the most efficient way to access and apply the colormap data.

        ```py
        geotiff = await GeoTIFF.open(...)
        array = await geotiff.fetch_tile(0, 0)

        colormap = geotiff.colormap
        colormap_array = colormap.as_array()

        rgb_data = colormap_array[array.data[0]]
        # A 3D array with shape (height, width, 3)
        ```

        Returns:
            A NumPy array representation of the colormap.

        """
        cmap_array = np.asarray(self._cmap)
        if dtype == np.uint8:
            return (cmap_array >> 8).astype(np.uint8)
        if dtype == np.uint16:
            return cmap_array
        raise ValueError("dtype must be either np.uint8 or np.uint16.")

    def as_dict(
        self,
        *,
        dtype: type[np.uint8 | np.uint16] = np.uint8,
    ) -> dict[int, tuple[int, int, int]]:
        """Return the colormap as a dictionary mapping indices to RGB tuples.

        Returns:
            A dictionary where keys are indices and values are tuples of
                (Red, Green, Blue) components.

        """
        cmap_array = self.as_array(dtype=dtype)
        return {
            int(idx): (int(r), int(g), int(b))
            for idx, (r, g, b) in enumerate(cmap_array)
        }

    def as_rasterio(self) -> dict[int, tuple[int, int, int, int]]:
        """Return the colormap as a mapping to 8-bit RGBA colors.

        This returns a colormap in the same format as rasterio's
        [`DatasetReader.colormap`][rasterio.io.DatasetReader.colormap] method.

        This is the same as
        [`Colormap.as_dict`][async_geotiff.colormap.Colormap.as_dict] with:

        - `dtype` set to `np.uint8`
        - an added alpha channel set to 255, **except** for the nodata value, if
          defined, which has an alpha of 0.

        Returns:
            Mapping of color index value (starting at 0) to RGBA color as a 4-element
                tuple.

        """
        cmap_array = self.as_array(dtype=np.uint8)
        cmap_dict: dict[int, tuple[int, int, int, int]] = {}

        for idx, (r, g, b) in enumerate(cmap_array):
            alpha = 255 if self._nodata is None or idx != self._nodata else 0
            cmap_dict[int(idx)] = (int(r), int(g), int(b), alpha)

        return cmap_dict



"""Parse a pyproj CRS from a GeoKeyDirectory.

In the simple case, a GeoTIFF is defined by an EPSG code, which we can pass to
`pyproj.CRS.from_epsg`.

In the complex case, a GeoTIFF is defined by a set of geo keys that specify all
the parameters of a user-defined CRS. In this case, we need to build a PROJJSON
dictionary representing the CRS, which we can then pass to `pyproj.CRS.from_json_dict`.

Most of the code to construct a custom PROJJSON is generated by Claude. For any files
where CRS generation fails, we should add a new test case in `test_crs.py`.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

from pyproj import CRS
from pyproj.crs import CoordinateOperation  # type: ignore[import]

if TYPE_CHECKING:
    from async_tiff import GeoKeyDirectory

MODEL_TYPE_PROJECTED = 1
"""GeoTIFF model type constant for projected coordinate systems."""

MODEL_TYPE_GEOGRAPHIC = 2
"""GeoTIFF model type constant for geographic coordinate systems."""

USER_DEFINED_CRS = 32767
"""Sentinel value for user-defined CRS."""

# The two "public" functions of this module
__all__ = ["crs_from_geo_keys", "projjson_from_geo_keys"]


def crs_from_geo_keys(gkd: GeoKeyDirectory) -> CRS:
    """Parse a pyproj CRS from a GeoKeyDirectory.

    Supports both EPSG-coded and user-defined CRSes.
    """
    model_type = gkd.model_type

    if model_type == MODEL_TYPE_PROJECTED:
        crs = _projected_projection(gkd)
        if isinstance(crs, int):
            return CRS.from_epsg(crs)

        return CRS.from_json_dict(crs)

    if model_type == MODEL_TYPE_GEOGRAPHIC:
        crs = _geographic_projection(gkd)
        if isinstance(crs, int):
            return CRS.from_epsg(crs)

        return CRS.from_json_dict(crs)

    raise ValueError(f"Unsupported GeoTIFF model type: {model_type}")


def projjson_from_geo_keys(gkd: GeoKeyDirectory) -> dict:
    """Build a PROJJSON dict from a GeoKeyDirectory.

    For EPSG-coded CRSes, resolves via pyproj and returns the canonical
    PROJJSON.

    For user-defined CRSes, constructs the PROJJSON from individual
    geo key parameters.
    """
    model_type = gkd.model_type

    if model_type == MODEL_TYPE_PROJECTED:
        crs = _projected_projection(gkd)
        if isinstance(crs, int):
            return CRS.from_epsg(crs).to_json_dict()

        return crs

    if model_type == MODEL_TYPE_GEOGRAPHIC:
        crs = _geographic_projection(gkd)
        if isinstance(crs, int):
            return CRS.from_epsg(crs).to_json_dict()

        return crs

    raise ValueError(f"Unsupported GeoTIFF model type: {model_type}")


def _geographic_projection(gkd: GeoKeyDirectory) -> int | dict:
    """Infer the geographic CRS from geo keys.

    If an EPSG code is present, returns that integer. Otherwise, constructs
    a user-defined geographic CRS PROJJSON dict.
    """
    epsg = gkd.geographic_type
    if epsg is not None and epsg != USER_DEFINED_CRS:
        return epsg

    return _build_user_defined_geographic_projjson(gkd)


def _projected_projection(gkd: GeoKeyDirectory) -> int | dict:
    """Infer the projected CRS from geo keys.

    If an EPSG code is present, returns that integer. Otherwise, constructs
    a user-defined projected CRS PROJJSON dict.
    """
    epsg = gkd.projected_type
    if epsg is not None and epsg != USER_DEFINED_CRS:
        return epsg

    return _build_user_defined_projected_projjson(gkd)


def _parse_geog_citation(citation: str) -> dict[str, str]:
    """Parse a GDAL-style geog_citation string into a dict of key/value pairs.

    Example input:
        "GCS Name = My CRS|Datum = My Datum|Primem = Greenwich|"
    """
    result: dict[str, str] = {}
    for part in citation.split("|"):
        stripped = part.strip()
        if "=" in stripped:
            k, v = stripped.split("=", 1)
            result[k.strip()] = v.strip()
    return result


def _build_user_defined_geographic_projjson(gkd: GeoKeyDirectory) -> dict:
    """Build a user-defined geographic CRS PROJJSON dict.

    Constructs PROJJSON from the ellipsoid, datum, prime meridian, and angular
    unit parameters stored in the GeoKeyDirectory.
    """
    # Build ellipsoid parameters
    ellipsoid = _build_ellipsoid_params(gkd)

    # Build prime meridian
    pm_name = "Greenwich"
    pm_longitude = 0.0
    if (
        gkd.geog_prime_meridian is not None
        and gkd.geog_prime_meridian != USER_DEFINED_CRS
    ):
        # Known prime meridian by EPSG code
        pm_name = f"EPSG:{gkd.geog_prime_meridian}"
    elif gkd.geog_prime_meridian_long is not None:
        pm_longitude = gkd.geog_prime_meridian_long
        pm_name = "User-defined"

    # Try to extract structured names from GDAL-style geog_citation
    citation_parts = (
        _parse_geog_citation(gkd.geog_citation) if gkd.geog_citation else {}
    )
    gcs_name = citation_parts.get("GCS Name") or gkd.geog_citation or "User-defined"
    datum_name = citation_parts.get("Datum") or gkd.geog_citation or "User-defined"
    if citation_parts.get("Primem"):
        pm_name = citation_parts["Primem"]

    # Build datum
    datum_json: dict
    if (
        gkd.geog_geodetic_datum is not None
        and gkd.geog_geodetic_datum != USER_DEFINED_CRS
    ):
        datum_json = {
            "type": "GeodeticReferenceFrame",
            "name": f"Unknown datum based upon EPSG {gkd.geog_geodetic_datum} ellipsoid",  # noqa: E501
        }
    else:
        datum_json = {
            "type": "GeodeticReferenceFrame",
            "name": datum_name,
            "ellipsoid": ellipsoid,
            "prime_meridian": {
                "name": pm_name,
                "longitude": pm_longitude,
            },
        }

    return {
        "type": "GeographicCRS",
        "$schema": "https://proj.org/schemas/v0.7/projjson.schema.json",
        "name": gcs_name,
        "datum": datum_json,
        "coordinate_system": _geographic_cs(gkd),
    }


def _build_user_defined_projected_projjson(gkd: GeoKeyDirectory) -> dict:
    """Build a user-defined projected CRS PROJJSON dict.

    Constructs PROJJSON from the geographic base CRS and projection parameters
    stored in the GeoKeyDirectory.
    """
    # Build the base geographic CRS as PROJJSON
    base_crs = _geographic_projection(gkd)
    if isinstance(base_crs, int):
        base_crs_json = CRS.from_epsg(base_crs).to_json_dict()
    else:
        base_crs_json = base_crs

    # Build the coordinate operation (projection)
    conversion = _build_conversion(gkd)

    # Build the coordinate system for the projected CRS
    cs = _projected_cs(gkd)

    return {
        "type": "ProjectedCRS",
        "$schema": "https://proj.org/schemas/v0.7/projjson.schema.json",
        "name": gkd.proj_citation or gkd.citation or "User-defined",
        "base_crs": base_crs_json,
        "conversion": conversion,
        "coordinate_system": cs,
    }


def _build_ellipsoid_params(gkd: GeoKeyDirectory) -> dict:
    """Build ellipsoid JSON parameters from geo keys."""
    if gkd.geog_ellipsoid is not None and gkd.geog_ellipsoid != USER_DEFINED_CRS:
        # Known ellipsoid by EPSG code — use parameters from geo keys if present,
        # and include the EPSG id so pyproj resolves the canonical name.
        ellipsoid: dict = {
            "name": f"EPSG ellipsoid {gkd.geog_ellipsoid}",
            "id": {"authority": "EPSG", "code": gkd.geog_ellipsoid},
        }

        if gkd.geog_semi_major_axis is not None:
            ellipsoid["semi_major_axis"] = gkd.geog_semi_major_axis
        if gkd.geog_inv_flattening is not None:
            ellipsoid["inverse_flattening"] = gkd.geog_inv_flattening
        elif gkd.geog_semi_minor_axis is not None:
            ellipsoid["semi_minor_axis"] = gkd.geog_semi_minor_axis

        return ellipsoid

    # Fully user-defined ellipsoid
    semi_major = gkd.geog_semi_major_axis
    if semi_major is None:
        raise ValueError("User-defined ellipsoid requires geog_semi_major_axis")

    user_ellipsoid: dict = {
        "name": "User-defined",
        "semi_major_axis": semi_major,
    }

    if gkd.geog_inv_flattening is not None:
        user_ellipsoid["inverse_flattening"] = gkd.geog_inv_flattening
    elif gkd.geog_semi_minor_axis is not None:
        user_ellipsoid["semi_minor_axis"] = gkd.geog_semi_minor_axis
    else:
        raise ValueError(
            "User-defined ellipsoid requires geog_inv_flattening or geog_semi_minor_axis",  # noqa: E501
        )

    return user_ellipsoid


# GeoTIFF coordinate transformation type codes (GeoKey 3075)
# Ref: http://geotiff.maptools.org/spec/geotiff6.html#6.3.3.3
CT_TRANSVERSE_MERCATOR = 1
CT_TRANSVERSE_MERCATOR_SOUTH = 2
CT_OBLIQUE_MERCATOR = 3
CT_OBLIQUE_MERCATOR_LABORDE = 4
CT_OBLIQUE_MERCATOR_ROSENMUND = 5
CT_OBLIQUE_MERCATOR_SPHERICAL = 6
CT_MERCATOR = 7
CT_LAMBERT_CONFORMAL_CONIC_2SP = 8
CT_LAMBERT_CONFORMAL_CONIC_1SP = 9
CT_LAMBERT_AZIMUTHAL_EQUAL_AREA = 10
CT_ALBERS_EQUAL_AREA = 11
CT_AZIMUTHAL_EQUIDISTANT = 12
CT_STEREOGRAPHIC = 14
CT_POLAR_STEREOGRAPHIC = 15
CT_OBLIQUE_STEREOGRAPHIC = 16
CT_EQUIRECTANGULAR = 17
CT_CASSINI_SOLDNER = 18
CT_ORTHOGRAPHIC = 21
CT_POLYCONIC = 22
CT_SINUSOIDAL = 24
CT_NEW_ZEALAND_MAP_GRID = 26
CT_TRANSVERSE_MERCATOR_SOUTH_ORIENTED = 27


def _build_conversion(gkd: GeoKeyDirectory) -> dict:  # noqa: C901, PLR0912, PLR0915
    """Build a PROJ JSON conversion (coordinate operation) from geo keys."""
    ct = gkd.proj_coord_trans
    if ct is None:
        # Fall back to the named projection code (GeoKey 3074) if available
        if gkd.projection is not None and gkd.projection != USER_DEFINED_CRS:
            conv = CoordinateOperation.from_epsg(gkd.projection).to_json_dict()
            conv.pop("$schema", None)
            return conv
        raise ValueError("User-defined projected CRS requires proj_coord_trans")

    # Helper to build a projection parameter with unit
    def _angular(name: str, value: float | None, default: float = 0.0) -> dict:
        return {
            "name": name,
            "value": value if value is not None else default,
            "unit": "degree",
        }

    def _linear(name: str, value: float | None, default: float = 0.0) -> dict:
        return {
            "name": name,
            "value": value if value is not None else default,
            "unit": "metre",
        }

    def _scale(name: str, value: float | None, default: float = 1.0) -> dict:
        return {
            "name": name,
            "value": value if value is not None else default,
            "unit": "unity",
        }

    name = "User-defined"
    method: dict
    parameters: list[dict]

    if ct == CT_TRANSVERSE_MERCATOR:
        name = "Transverse Mercator"
        method = {"name": "Transverse Mercator"}
        parameters = [
            _angular("Latitude of natural origin", gkd.proj_nat_origin_lat),
            _angular("Longitude of natural origin", gkd.proj_nat_origin_long),
            _scale("Scale factor at natural origin", gkd.proj_scale_at_nat_origin),
            _linear("False easting", gkd.proj_false_easting),
            _linear("False northing", gkd.proj_false_northing),
        ]

    elif ct == CT_TRANSVERSE_MERCATOR_SOUTH:
        name = "Transverse Mercator (South Orientated)"
        method = {"name": "Transverse Mercator (South Orientated)"}
        parameters = [
            _angular("Latitude of natural origin", gkd.proj_nat_origin_lat),
            _angular("Longitude of natural origin", gkd.proj_nat_origin_long),
            _scale("Scale factor at natural origin", gkd.proj_scale_at_nat_origin),
            _linear("False easting", gkd.proj_false_easting),
            _linear("False northing", gkd.proj_false_northing),
        ]

    elif ct in (
        CT_OBLIQUE_MERCATOR,
        CT_OBLIQUE_MERCATOR_LABORDE,
        CT_OBLIQUE_MERCATOR_ROSENMUND,
        CT_OBLIQUE_MERCATOR_SPHERICAL,
    ):
        name = "Hotine Oblique Mercator (variant B)"
        method = {"name": "Hotine Oblique Mercator (variant B)"}
        parameters = [
            _angular("Latitude of projection centre", gkd.proj_center_lat),
            _angular("Longitude of projection centre", gkd.proj_center_long),
            _angular("Azimuth of initial line", gkd.proj_azimuth_angle),
            _angular("Angle from Rectified to Skew Grid", gkd.proj_azimuth_angle),
            _scale("Scale factor on initial line", gkd.proj_scale_at_center),
            _linear("Easting at projection centre", gkd.proj_center_easting),
            _linear("Northing at projection centre", gkd.proj_center_northing),
        ]

    elif ct == CT_MERCATOR:
        name = "Mercator (variant A)"
        method = {"name": "Mercator (variant A)"}
        parameters = [
            _angular("Latitude of natural origin", gkd.proj_nat_origin_lat),
            _angular("Longitude of natural origin", gkd.proj_nat_origin_long),
            _scale("Scale factor at natural origin", gkd.proj_scale_at_nat_origin),
            _linear("False easting", gkd.proj_false_easting),
            _linear("False northing", gkd.proj_false_northing),
        ]

    elif ct == CT_LAMBERT_CONFORMAL_CONIC_2SP:
        name = "Lambert Conic Conformal (2SP)"
        method = {"name": "Lambert Conic Conformal (2SP)"}
        parameters = [
            _angular(
                "Latitude of false origin",
                gkd.proj_false_origin_lat or gkd.proj_nat_origin_lat,
            ),
            _angular(
                "Longitude of false origin",
                gkd.proj_false_origin_long or gkd.proj_nat_origin_long,
            ),
            _angular("Latitude of 1st standard parallel", gkd.proj_std_parallel1),
            _angular("Latitude of 2nd standard parallel", gkd.proj_std_parallel2),
            _linear(
                "Easting at false origin",
                gkd.proj_false_origin_easting or gkd.proj_false_easting,
            ),
            _linear(
                "Northing at false origin",
                gkd.proj_false_origin_northing or gkd.proj_false_northing,
            ),
        ]

    elif ct == CT_LAMBERT_CONFORMAL_CONIC_1SP:
        name = "Lambert Conic Conformal (1SP)"
        method = {"name": "Lambert Conic Conformal (1SP)"}
        parameters = [
            _angular("Latitude of natural origin", gkd.proj_nat_origin_lat),
            _angular("Longitude of natural origin", gkd.proj_nat_origin_long),
            _scale("Scale factor at natural origin", gkd.proj_scale_at_nat_origin),
            _linear("False easting", gkd.proj_false_easting),
            _linear("False northing", gkd.proj_false_northing),
        ]

    elif ct == CT_LAMBERT_AZIMUTHAL_EQUAL_AREA:
        name = "Lambert Azimuthal Equal Area"
        method = {"name": "Lambert Azimuthal Equal Area"}
        parameters = [
            _angular("Latitude of natural origin", gkd.proj_center_lat),
            _angular("Longitude of natural origin", gkd.proj_center_long),
            _linear("False easting", gkd.proj_false_easting),
            _linear("False northing", gkd.proj_false_northing),
        ]

    elif ct == CT_ALBERS_EQUAL_AREA:
        name = "Albers Equal Area"
        method = {"name": "Albers Equal Area"}
        parameters = [
            _angular(
                "Latitude of false origin",
                gkd.proj_false_origin_lat or gkd.proj_nat_origin_lat,
            ),
            _angular(
                "Longitude of false origin",
                gkd.proj_false_origin_long or gkd.proj_nat_origin_long,
            ),
            _angular("Latitude of 1st standard parallel", gkd.proj_std_parallel1),
            _angular("Latitude of 2nd standard parallel", gkd.proj_std_parallel2),
            _linear(
                "Easting at false origin",
                gkd.proj_false_origin_easting or gkd.proj_false_easting,
            ),
            _linear(
                "Northing at false origin",
                gkd.proj_false_origin_northing or gkd.proj_false_northing,
            ),
        ]

    elif ct == CT_AZIMUTHAL_EQUIDISTANT:
        name = "Modified Azimuthal Equidistant"
        method = {"name": "Modified Azimuthal Equidistant"}
        parameters = [
            _angular("Latitude of natural origin", gkd.proj_center_lat),
            _angular("Longitude of natural origin", gkd.proj_center_long),
            _linear("False easting", gkd.proj_false_easting),
            _linear("False northing", gkd.proj_false_northing),
        ]

    elif ct == CT_STEREOGRAPHIC:
        name = "Stereographic"
        method = {"name": "Stereographic"}
        parameters = [
            _angular("Latitude of natural origin", gkd.proj_center_lat),
            _angular("Longitude of natural origin", gkd.proj_center_long),
            _scale("Scale factor at natural origin", gkd.proj_scale_at_center),
            _linear("False easting", gkd.proj_false_easting),
            _linear("False northing", gkd.proj_false_northing),
        ]

    elif ct == CT_POLAR_STEREOGRAPHIC:
        name = "Polar Stereographic (variant B)"
        method = {"name": "Polar Stereographic (variant B)"}
        parameters = [
            _angular(
                "Latitude of standard parallel",
                gkd.proj_nat_origin_lat or gkd.proj_std_parallel1,
            ),
            _angular(
                "Longitude of origin",
                gkd.proj_straight_vert_pole_long or gkd.proj_nat_origin_long,
            ),
            _linear("False easting", gkd.proj_false_easting),
            _linear("False northing", gkd.proj_false_northing),
        ]

    elif ct == CT_OBLIQUE_STEREOGRAPHIC:
        name = "Oblique Stereographic"
        method = {"name": "Oblique Stereographic"}
        parameters = [
            _angular("Latitude of natural origin", gkd.proj_center_lat),
            _angular("Longitude of natural origin", gkd.proj_center_long),
            _scale("Scale factor at natural origin", gkd.proj_scale_at_center),
            _linear("False easting", gkd.proj_false_easting),
            _linear("False northing", gkd.proj_false_northing),
        ]

    elif ct == CT_EQUIRECTANGULAR:
        name = "Equidistant Cylindrical"
        method = {"name": "Equidistant Cylindrical"}
        parameters = [
            _angular(
                "Latitude of 1st standard parallel",
                gkd.proj_std_parallel1 or gkd.proj_center_lat,
            ),
            _angular("Longitude of natural origin", gkd.proj_center_long),
            _linear("False easting", gkd.proj_false_easting),
            _linear("False northing", gkd.proj_false_northing),
        ]

    elif ct == CT_CASSINI_SOLDNER:
        name = "Cassini-Soldner"
        method = {"name": "Cassini-Soldner"}
        parameters = [
            _angular("Latitude of natural origin", gkd.proj_nat_origin_lat),
            _angular("Longitude of natural origin", gkd.proj_nat_origin_long),
            _linear("False easting", gkd.proj_false_easting),
            _linear("False northing", gkd.proj_false_northing),
        ]

    elif ct == CT_POLYCONIC:
        name = "American Polyconic"
        method = {"name": "American Polyconic"}
        parameters = [
            _angular("Latitude of natural origin", gkd.proj_nat_origin_lat),
            _angular("Longitude of natural origin", gkd.proj_nat_origin_long),
            _linear("False easting", gkd.proj_false_easting),
            _linear("False northing", gkd.proj_false_northing),
        ]

    elif ct == CT_SINUSOIDAL:
        name = "Sinusoidal"
        method = {"name": "Sinusoidal"}
        parameters = [
            _angular("Longitude of natural origin", gkd.proj_center_long),
            _linear("False easting", gkd.proj_false_easting),
            _linear("False northing", gkd.proj_false_northing),
        ]

    elif ct == CT_ORTHOGRAPHIC:
        name = "Orthographic"
        method = {"name": "Orthographic"}
        parameters = [
            _angular("Latitude of natural origin", gkd.proj_center_lat),
            _angular("Longitude of natural origin", gkd.proj_center_long),
            _linear("False easting", gkd.proj_false_easting),
            _linear("False northing", gkd.proj_false_northing),
        ]

    elif ct == CT_NEW_ZEALAND_MAP_GRID:
        name = "New Zealand Map Grid"
        method = {"name": "New Zealand Map Grid"}
        parameters = [
            _angular("Latitude of natural origin", gkd.proj_nat_origin_lat),
            _angular("Longitude of natural origin", gkd.proj_nat_origin_long),
            _linear("False easting", gkd.proj_false_easting),
            _linear("False northing", gkd.proj_false_northing),
        ]

    elif ct == CT_TRANSVERSE_MERCATOR_SOUTH_ORIENTED:
        name = "Transverse Mercator (South Orientated)"
        method = {"name": "Transverse Mercator (South Orientated)"}
        parameters = [
            _angular("Latitude of natural origin", gkd.proj_nat_origin_lat),
            _angular("Longitude of natural origin", gkd.proj_nat_origin_long),
            _scale("Scale factor at natural origin", gkd.proj_scale_at_nat_origin),
            _linear("False easting", gkd.proj_false_easting),
            _linear("False northing", gkd.proj_false_northing),
        ]

    else:
        raise ValueError(f"Unsupported coordinate transformation type: {ct}")

    return {
        "name": name,
        "method": method,
        "parameters": parameters,
    }


ANGULAR_UNIT_RADIAN = 9101
"""EPSG code for radian angular unit."""

ANGULAR_UNIT_DEGREE = 9102
"""EPSG code for degree angular unit."""

ANGULAR_UNIT_GRAD = 9105
"""EPSG code for grad angular unit."""

ANGULAR_UNIT: dict[int | None, str] = {
    ANGULAR_UNIT_DEGREE: "degree",
    ANGULAR_UNIT_RADIAN: "radian",
    ANGULAR_UNIT_GRAD: "grad",
}


def _geographic_cs(gkd: GeoKeyDirectory) -> dict:
    """Build a geographic coordinate system JSON dict from geo keys."""
    angular_unit = ANGULAR_UNIT.get(
        gkd.geog_angular_units,
        "degree",
    )

    return {
        "subtype": "ellipsoidal",
        "axis": [
            {
                "name": "Latitude",
                "abbreviation": "lat",
                "direction": "north",
                "unit": angular_unit,
            },
            {
                "name": "Longitude",
                "abbreviation": "lon",
                "direction": "east",
                "unit": angular_unit,
            },
        ],
    }


LINEAR_UNIT_METRE = 9001
LINEAR_UNIT_FOOT = 9002
LINEAR_UNIT_US_SURVEY_FOOT = 9003

LINEAR_UNIT: dict[int | None, str | dict[str, str | float]] = {
    LINEAR_UNIT_METRE: "metre",
    LINEAR_UNIT_FOOT: "foot",
    LINEAR_UNIT_US_SURVEY_FOOT: {
        "type": "LinearUnit",
        "name": "US survey foot",
        "conversion_factor": 0.30480060960121924,
    },
}


def _projected_cs(gkd: GeoKeyDirectory) -> dict:
    """Build a projected coordinate system JSON dict from geo keys."""
    linear_unit: str | dict[str, str | float] = LINEAR_UNIT.get(
        gkd.proj_linear_units,
        "metre",
    )

    return {
        "subtype": "Cartesian",
        "axis": [
            {
                "name": "Easting",
                "abbreviation": "E",
                "direction": "east",
                "unit": linear_unit,
            },
            {
                "name": "Northing",
                "abbreviation": "N",
                "direction": "north",
                "unit": linear_unit,
            },
        ],
    }
